import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:build/build.dart';
import 'package:fireproof/src/utils.dart';
import 'package:source_gen/source_gen.dart';
import 'package:strings/strings.dart';

class CollectionOutput {
  /// The string name of the collection the firestore database
  final String collectionName;

  final String collectionClassName;

  final String fieldName;

  final String output;

  final String docName;

  final String modelName;

  const CollectionOutput({
    required this.collectionName,
    required this.collectionClassName,
    required this.fieldName,
    required this.output,
    required this.docName,
    required this.modelName,
  });
}

Future<CollectionOutput> collectionGenerator(
  PropertyAccessorElement element,
  BuildStep buildStep,
) async {
  final stringBuffer = StringBuffer();
  final displayName = element.displayName;
  final collectionType = element.returnType;
  final collectionElement = collectionType.element;

  if (collectionElement == null) {
    throw InvalidGenerationSourceError(
      'Collections must have a valid type of FireCollectionReference. Failing '
      'element: ${element.name}',
      element: element,
    );
  }

  final collectionSuperType = collectionElement as ClassElement;
  final collectionSuperTypeName = collectionSuperType.name;
  final node = await tryGetAstNodeForElement(element, buildStep);
  final collectionTypes = getGenericTypes(node);

  if (collectionType is! InterfaceType ||
      collectionSuperTypeName != 'FireCollectionReference') {
    throw InvalidGenerationSourceError(
      'Collections must have a valid type of FireCollectionReference. Failing '
      'element: ${element.name}',
      element: element,
    );
  }

  if (collectionTypes.length != 2) {
    throw InvalidGenerationSourceError(
      'Type FireCollectionReference must be declared with two types <Doc, '
      'Model>. Failing element: ${element.name}',
      element: element,
    );
  }

  // TODO: This will likely not be generated in time
  // final docName =
  //     collectionTypes.first.getDisplayString(withNullability: false);

  final modelName = collectionTypes.first.lexeme;
  final docName = collectionTypes.last.lexeme;
  final lowerCaseModel = lowerTitleCase(modelName);
  final suffix = 'Collection';
  final className = capitalize(displayName) + suffix;
  final querySnapshotName = capitalize(displayName) + 'QuerySnapshot';
  final queryName = capitalize(displayName) + 'Query';

  stringBuffer.write('''
class $className extends $queryName with FireCollectionReference<$modelName, $docName> {
  final CollectionReference _reference;

  $className(this._reference) : super(_reference);

  @override
  FirebaseFirestore get firestore => _reference.firestore;

  @override
  String get id => _reference.id;

  @override
  FireCollectionReference? get parent => throw UnimplementedError();

  @override
  String get path => _reference.path;

  @override
  Future<$docName> add($modelName $lowerCaseModel) async {
    final docReference = await _reference.add($lowerCaseModel.toJson());

    return $docName(docReference);
  }

  @override
  $docName doc([String? path]) {
    final documentReference = _reference.doc(path);

    return $docName(documentReference);
  }

  @override
  Stream<$querySnapshotName> snapshots({bool includeMetadataChanges = false}) {
    return _reference
        .snapshots(includeMetadataChanges: includeMetadataChanges)
        .map((snapshot) => $querySnapshotName(snapshot));
  }

  @override
  Future<$querySnapshotName> get([GetOptions? getOptions]) async {
    final snapshot = await _reference.get(getOptions);

    return $querySnapshotName(snapshot);
  }
}

''');

  stringBuffer.write('''
class $queryName extends FireQuery<$modelName, $docName> {
  final Query _query;

  $queryName(this._query);

  @override
  FirebaseFirestore get firestore => _query.firestore;

  @override
  Map<String, dynamic> get parameters {
    return _query.parameters;
  }

  @override
  $queryName endAtDocument(DocumentSnapshot documentSnapshot) {
    // TODO: Typesafe FireDocumentSnapshot should be used
    final query = _query.endAtDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName endAt(List values) {
    final query = _query.endAt(values);

    return $queryName(query);
  }

  @override
  $queryName endBeforeDocument(DocumentSnapshot documentSnapshot) {
    // TODO: Typesafe FireDocumentSnapshot should be used
    final query = _query.endBeforeDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName endBefore(List values) {
    final query = _query.endBefore(values);

    return $queryName(query);
  }

  @override
  Future<$querySnapshotName> get([GetOptions? options]) async {
    final querySnapshot = await _query.get(options);

    return $querySnapshotName(querySnapshot);
  }

  @override
  $queryName limit(int limit) {
    final query = _query.limit(limit);

    return $queryName(query);
  }

  @override
  $queryName limitToLast(int limit) {
    final query = _query.limit(limit);

    return $queryName(query);
  }

  @override
  Stream<$querySnapshotName> snapshots(
      {bool includeMetadataChanges = false}){
    throw UnimplementedError();
  }

  @override
  $queryName orderBy(dynamic field, {bool descending = false}) {
    final query = _query.orderBy(field, descending: descending);

    return $queryName(query);
  }

  @override
  $queryName startAfterDocument(DocumentSnapshot documentSnapshot) {
    // TODO: Typesafe FireDocumentSnapshot should be used
    final query = _query.startAfterDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName startAfter(List values) {
    final query = _query.startAfter(values);

    return $queryName(query);
  }

  @override
  $queryName startAtDocument(DocumentSnapshot documentSnapshot) {
    // TODO: Typesafe FireDocumentSnapshot should be used
    final query = _query.startAtDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName startAt(List values) {
    final query = _query.startAt(values);

    return $queryName(query);
  }

  @override
  $queryName where(dynamic field,
      {dynamic isEqualTo,
      dynamic isNotEqualTo,
      dynamic isLessThan,
      dynamic isLessThanOrEqualTo,
      dynamic isGreaterThan,
      dynamic isGreaterThanOrEqualTo,
      dynamic arrayContains,
      List? arrayContainsAny,
      List? whereIn,
      List? whereNotIn,
      bool? isNull}){
    throw UnimplementedError();
  }
}
''');

  stringBuffer.write('''
class $querySnapshotName with FireQuerySnapshot<$modelName, $docName> {
  final QuerySnapshot _snapshot;

  const $querySnapshotName(this._snapshot);

  @override
  List<$queryName> get docs {
    throw UnimplementedError();
  }

  @override
  List<FireDocumentChange<$modelName, $docName>> get docChanges {
    throw UnimplementedError();
  }

  @override
  SnapshotMetadata get metadata => _snapshot.metadata;
}

''');

  return CollectionOutput(
    fieldName: displayName,
    output: stringBuffer.toString(),
    collectionName: element.displayName,
    collectionClassName: className,
    modelName: modelName,
    docName: docName,
  );
}
