import 'dart:async';

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:fireproof_annotation/fireproof_annotation.dart'
    show Collection, Root;
import 'package:source_gen/source_gen.dart';

import 'collection_generator.dart';

class RootGenerator extends GeneratorForAnnotation<Root> {
  @override
  FutureOr<String> generateForAnnotatedElement(
    Element element,
    ConstantReader annotation,
    BuildStep buildStep,
  ) async {
    final stringBuffer = StringBuffer();

    if (element is! ClassElement) {
      throw InvalidGenerationSourceError(
        '@root can only be applied on classes. Failing element: ${element.name}',
        element: element,
      );
    }

    final collectionChecker = TypeChecker.fromRuntime(Collection);
    final collectionGetters = element.accessors
        .where((accessor) => accessor.isGetter)
        .where((element) => collectionChecker.hasAnnotationOf(element));

    if (!element.isAbstract && collectionGetters.isNotEmpty) {
      throw InvalidGenerationSourceError(
        '@fireproof classes that declare collections must be marked abstract. '
        'Failing element: ${element.name}',
        element: element,
      );
    }

    final collectionOutputFutures = collectionGetters
        .map<Future<CollectionOutput>>(
            (element) => collectionGenerator(element, buildStep))
        .toList();
    final collectionOutputs = await Future.wait(collectionOutputFutures);
    final displayName = element.displayName;
    final prefix = '_';
    final className = prefix + displayName;
    final initializerList = collectionOutputs.map((output) {
      return '${output.fieldName} = ${output.collectionClassName}(instance.collection(\'${output.collectionName}\'))';
    }).join(',');
    final initializerListString =
        initializerList.isEmpty ? '' : ' : $initializerList';
    final collectionFields = collectionOutputs.map((output) {
      return '''@override
  final ${output.collectionClassName} ${output.fieldName};''';
    }).join('\n\n\t');

    stringBuffer.write('''
class $className implements $displayName {
  $className(this.instance)$initializerListString;
  
  $collectionFields

  final FirebaseFirestore instance ;
}

''');

    for (final collectionOutput in collectionOutputs) {
      stringBuffer.write(collectionOutput.output);
    }

    return stringBuffer.toString();
  }
}
